( cd usr/doc/svgalib-1.4.3/doc ; rm -rf README.lrmi )
( cd usr/doc/svgalib-1.4.3/doc ; ln -sf ../lrmi-0.6m/README README.lrmi )
( cd usr/lib ; rm -rf libvga.so )
( cd usr/lib ; ln -sf libvga.so.1.4.3 libvga.so )
( cd usr/lib ; rm -rf libvga.so.1 )
( cd usr/lib ; ln -sf libvga.so.1.4.3 libvga.so.1 )
( cd usr/lib ; rm -rf libvgagl.so )
( cd usr/lib ; ln -sf libvgagl.so.1.4.3 libvgagl.so )
( cd usr/lib ; rm -rf libvgagl.so.1 )
( cd usr/lib ; ln -sf libvgagl.so.1.4.3 libvgagl.so.1 )
